#!/usr/bin/env python3

import xml.etree.ElementTree as ET
import os
import sys
import logging
from hashlib import sha256

def pretty_xml(element, indent, newline, level=0):
    if element:
        if (element.text is None) or element.text.isspace():
            element.text = newline + indent * (level + 1)
        else:
            element.text = newline + indent * (level + 1) + element.text.strip() + newline + indent * (level + 1)
    temp = list(element)
    for subelement in temp:
        if temp.index(subelement) < (len(temp) - 1):
            subelement.tail = newline + indent * (level + 1)
        else:
            subelement.tail = newline + indent * level
        pretty_xml(subelement, indent, newline, level=level + 1)

class Object(object):

    def __init__(self):
        self.type = None
        self.name = None
        self.mode = None
        self.sha256 = None
        self.set = set()

    def to_xml_node(self):
        pass

    def from_xml_node(self, node):
        assert "object" == node.tag, "\033[1;31m[ERROR]\033[0m Wrong node to parse as file."
        assert "type" in node.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"type\" attribute for file."
        assert "name" in node.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"name\" attribute for file."
        assert "mode" in node.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"mode\" attribute for file."
        assert "form" in node.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"form\" attribute for file."
        self.type = node.attrib["type"]
        self.name = node.attrib["name"]
        self.mode = node.attrib["mode"]
        self.form = node.attrib["form"]
        if "sha256" in node.attrib:
            self.sha256 = node.attrib["sha256"]
        if "set" in node.attrib:
            self.set = set(node.attrib["set"].split(" "))

    def copy_from(self, other):
        self.type = other.type
        self.name = other.name
        self.mode = other.mode
        self.form = other.form
        self.sha256 = other.sha256

class Manifest(object):

    def __init__(self):
        self.objects = dict()
        self.valid_sets = None
        self.tree = None
        self.version = None
        self.is_valid_ = False

    def parse_from_file(self, manifest_file):
        if not os.path.exists(manifest_file):
            print("\033[1;31m[ERROR]\033[0m Parsing manifest failed. File does not exist: {}"
                    .format(manifest_file), file=sys.stderr)
            sys.exit(1)
        self.tree = ET.parse(manifest_file)
        root = self.tree.getroot()
        if not root.tag == "binary_release_manifest":
            print("\033[1;31m[ERROR]\033[0m Parsing manifest failed. File does not have a binary_release_manifest tag: {}"
                    .format(manifest_file), file=sys.stderr)
            sys.exit(1)

        for child in root:
            if child.tag == "general":
                assert "version" in child.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"version\" info"
                self.version = child.attrib["version"]
            if child.tag == "valid_sets":
                self.valid_sets = set(child.text.split(" "))
            if child.tag == "object":
                current_object = Object()
                current_object.from_xml_node(child)
                if current_object.type in self.objects:
                    self.objects[current_object.type].append(current_object)
                else:
                    self.objects[current_object.type] = [current_object]
        if self.check_objects_validity():
            self.is_valid_ = True
        else:
            self.is_valid_ = False

    def is_valid(self):
        return self.is_valid_

    def change_to_release_and_add_sha256(self,name,sha256):
        root = self.tree.getroot()
        for child in root:
            if child.tag == "object" and child.attrib["name"] == name:
                if child.attrib["mode"] == "test":
                    child.set("mode","release")
                    child.set("sha256",sha256)

    def format_xml(self,manifest_file):
        with open(manifest_file) as f:
            content = f.read()
            content = content.replace(" sha256=","\n\t\tsha256=")
            content = content.replace(" mode=","\n\t\tmode=")
            content = content.replace(" name=","\n\t\tname=")
            content = content.replace(" type=","\n\t\ttype=")
            content = content.replace(" form=","\n\t\tform=")
            content = content.replace(" desc=","\n\t\tdesc=")
            content = content.replace(" set=","\n\t\tset=")
        with open(manifest_file, mode="w") as w:
            w.write(content)

    def generate_to_file(self,manifest_file):
        root = self.tree.getroot()
        for child in root:
            if child.tag == "general":
                assert "version" in child.attrib, "\033[1;31m[ERROR]\033[0m Does not have \"version\" info"
                # add a version number
                child.set("version",self.raise_version())
        pretty_xml(root, '\t', '\n')
        ret = self.tree.write("{}".format(manifest_file))
        self.format_xml(manifest_file)
        return ret

    def raise_version(self):
        v = self.version
        v = v[:v.rfind(".")]+"."+str(int(v.split(".")[-1])+1)
        return v

    def check_objects_validity(self):
        # objects with same type should not be in the same set
        for obj_type, objects in self.objects.items():
            union_set = set()
            for obj in objects:
                for set_item in obj.set:
                    if set_item in union_set:
                        logging.error("Got two objects with same type {} in the same set {}".format(obj_type, set_item))
                        return False
                    elif self.valid_sets and not (set_item in self.valid_sets):
                        logging.error("Invalid set {} for item {}".format(set_item, obj_type))
                        logging.error("Valid sets : {}".format(str(self.valid_sets)))
                        return False
                    else:
                        union_set.add(set_item)

        return True

if __name__ == "__main__":
    manifest_path = "./model_manifest.xml"
    manifest = Manifest()
    manifest.parse_from_file(manifest_path)
    print("Whether manifest is valid : {}".format(manifest.is_valid()))
    print("Valid set : {}".format(manifest.valid_sets))
    manifest.generate_to_file("./model_manifest_out.xml")
