#!/usr/bin/env python3

"""
provide Querier for binary release remote object management.
Querier class is in charge of :
   list remote objects
   upload object to aws
   download object from aws
"""

import os
import sys
import boto3
import argparse
import logging
import datetime
from dateutil.tz import tzutc

from .manifest_parser_and_generator import Object, Manifest
from .config_reader import ConfigReader, ModeConfig
from .path_handler import PathHandler
from .get_aws_endpoint import get_aws_endpoint

class RemoteObject(object):

    def __init__(self):
        self.name = None
        self.mode = None
        self.form = None
        self.remote_path = None
        self.last_modified = None
        self.commit_id = None

class Querier(object):

    def __init__(self, config_file):
        try:
            boto3.setup_default_session(profile_name="ad_system_common")
        except:
            print("\033[1;31m[ERROR]\033[0m Do not have aws profile 'ad_system_common'",
                  file=sys.stderr)

        self.config = ConfigReader()
        self.config.read(config_file)
        self.path_handler = PathHandler()
        self.remote_release_location = self.path_handler.ConfigurePath(self.config.modes["release"].remote_location)
        self.remote_test_location = self.path_handler.ConfigurePath(self.config.modes["test"].remote_location)
        self.aws_endpoint_url = get_aws_endpoint()
        self.s3_ = boto3.resource(service_name="s3", endpoint_url=self.aws_endpoint_url)
        self.check_access()
        self.remote_objects = dict()
        self.list_remote_objects()

    def check_access(self):
        # check if aws is installed
        if os.system("which aws >/dev/null 2>&1") != 0:
            logging.error("\033[1;31m[ERROR]\033[0m Make sure aws is installed."
                  "  You may install it using 'apt install awscli'")
            sys.exit(1)
        # check the access to ad_binary_release
        access = True
        try:
            null = self.s3_.meta.client.list_objects(
                Bucket="ad_binary_release",
                Prefix=""
            )
        except:
            access = False
        if access:
            logging.info("[CHECK_ACCESS] DONE.")
        else:
            logging.error("\033[1;31m[ERROR]\033[0m Do not have access to ad_binary_release.")
            print("  Please check the following items: \n"
                  "  1. you have configured ad_system_common profile for aws.\n"
                  "     If true, you should have corresponed item in ~/.aws/config.\n"
                  "  2. you have access to internal network of sw-motion.\n"
                  "  3. your access key and secret key are correct.\n"
                  "  You can retry with command 'aws s3 --endpoint_url={endpoint} --profile ad_system_common ls'".format(endpoint=self.aws_endpoint_url))

    def list_remote_objects(self):
        response_release = os.popen("aws s3 --endpoint-url={endpoint} --profile ad_system_common ls --recursive --page-size 4000 s3://{remote_path}".format(
            endpoint=self.aws_endpoint_url, remote_path=self.remote_release_location))
        for line in response_release.readlines():
            line = line.strip("\n")
            current_object = RemoteObject()
            current_object.mode = "release"
            current_object.name = line.split('/')[-1]
            current_object.remote_path = "s3://ad_binary_release/{}".format(line.split(' ')[-1])
            current_object.last_modified = datetime.datetime.strptime(line[:19], "%Y-%m-%d %H:%M:%S")
            if current_object.name in self.remote_objects.keys():
                self.remote_objects[current_object.name].append(current_object)
            else:
                self.remote_objects[current_object.name] = [current_object]

        response_test = os.popen("aws s3 --endpoint-url={endpoint} --profile ad_system_common ls --recursive --page-size 4000 s3://{remote_path}".format(
            endpoint=self.aws_endpoint_url, remote_path=self.remote_test_location))
        for line in response_test.readlines():
            line = line.strip("\n")
            current_object = RemoteObject()
            current_object.mode = "test"
            # judge if end with a commit id
            last_fragment = line.split(' ')[-1].split(".")[-1]
            tmp_list = [(ord(last_fragment[i]) - ord('a')) * \
                        (ord(last_fragment[i]) - ord('f')) * \
                        (ord(last_fragment[i]) - ord('0')) * \
                        (ord(last_fragment[i]) - ord('9')) <= 0 for i in range(len(last_fragment))]
            if len(last_fragment) == sum(tmp_list) and len(last_fragment) >= 6:
                current_object.name = line.split("/")[-1][:-1*len(last_fragment)-1]
                current_object.commit_id = last_fragment
            else:
                current_object.name = line.split("/")[-1]
            current_object.remote_path = "s3://ad_binary_release/{}".format(line.split(' ')[-1])
            current_object.last_modified = datetime.datetime.strptime(line[:19], "%Y-%m-%d %H:%M:%S")
            if current_object.name in self.remote_objects.keys():
                self.remote_objects[current_object.name].append(current_object)
            else:
                self.remote_objects[current_object.name] = [current_object]

    def find_proper_test_object(self, object_name, commit_id=None):
        """
        find the test object with the corresponding commit-id
        if not found, return the test object that is uploaded latest
        """
        if commit_id:
            remote_path = ""
            for remote_object in self.remote_objects[object_name]:
                if remote_object.commit_id == commit_id:
                    remote_path = remote_object.remote_path
                    break
            if not remote_path:
                logging.warn("\033[1;33m[WARN]\033[0m Can not find test object with specific commit id.")
        if not commit_id or not remote_path:
            latest_object = None
            for remote_object in self.remote_objects[object_name]:
                if remote_object.mode == "test":
                    latest_object = remote_object
                    break
            for remote_object in self.remote_objects[object_name]:
                if remote_object.mode == "test" and remote_object.last_modified > latest_object.last_modified:
                    latest_object = remote_object
            if not latest_object:
                logging.error("\033[1;31m[ERROR]\033[0m Test object \033[1;31m{}\033[0m not found.".format(object_name))
                for remote_object in self.remote_objects[object_name]:
                    logging.debug(remote_object.name," ",remote_object.mode," ", remote_object.remote_path)
                sys.exit(1)
            remote_path = latest_object.remote_path
            logging.info("remote path for {object} : {path}".format(object=object_name,path=remote_path))
        return remote_path

    def download_object(self, object_name, local_path, form, commit_id=None, object_mode="release"):
        """
        download a single object from AWS
        """
        if not object_name in self.remote_objects.keys():
            logging.error("\033[1;31m[ERROR]\033[0m Object {} not found from AWS.".format(object_name))
            return 1
        else:
            remote_path = None
            if object_mode == "test":
                remote_path = self.find_proper_test_object(object_name, commit_id)
            if object_mode == "release":
                for remote_object in self.remote_objects[object_name]:
                    if remote_object.mode == "release":
                        remote_path = remote_object.remote_path
                        break
                if not remote_path:
                    logging.error("\033[1;31m[ERROR]\033[0m Release object not found.")
                    return 1
        if form == "folder":
            cmd = "aws --endpoint-url={endpoint} --profile ad_system_common "\
                "s3 cp {remote_path} {local_path}/{name}.tar 2>/dev/null" \
                .format(endpoint=self.aws_endpoint_url, remote_path=remote_path, local_path=local_path, name=object_name)
        else:
            cmd = "aws --endpoint-url={endpoint} --profile ad_system_common "\
                "s3 cp {remote_path} {local_path}/{name} 2>/dev/null" \
                .format(endpoint=self.aws_endpoint_url, remote_path=remote_path, local_path=local_path, name=object_name)
        ret = os.system(cmd)
        if ret != 0:
            ret = os.system("sudo "+cmd)
            if ret != 0:
                logging.error("\033[1;31m[ERROR]\033[0m Failed while downloading object {name} with command {cmd}"
                    .format(name=object_name, cmd=cmd))

        if form == "folder":
            cmd = "tar -xf {local_path}/{name}.tar -C {local_path} >/dev/null 2>&1".format(local_path=local_path, name=object_name)
            ret = os.system(cmd)
            if 0 != ret:
                ret = os.system("sudo "+cmd)
                if 0 != ret:
                    logging.error("\033[1;31m[ERROR]\033[0m Un-tar {local_path}/{name}.tar failed.".format(local_path=local_path, name=object_name))
                else:
                    os.system("sudo rm {local_path}/{name}.tar".format(local_path=local_path, name=object_name))
            else:
                os.system("rm {local_path}/{name}.tar".format(local_path=local_path, name=object_name))

        return ret
        #self.s3_.meta.client.download_file('ad_binary_release', remote_file, path+"/"+model_name)

    def remove_all_test_objects(self, name):
        if name in self.remote_objects.keys():
            for target_object in self.remote_objects[name]:
                if target_object.mode == "test":
                    cmd = "aws --endpoint-url={endpoint} --profile ad_system_admin "\
                        "s3 mv {origin} {dest}".format(endpoint=self.aws_endpoint_url, origin=target_object.remote_path, dest=target_object.remote_path.replace("/test/","/archived/"))
                    os.system(cmd)

    def upload_object(self, file_local_name, file_remote_name):
        """
        upload a single test object to AWS
        """
        # prepare to a tmp tar
        # upload
        cmd = "aws --endpoint-url={endpoint} --profile ad_system_common "\
              "s3 cp {local_name} s3://{remote_path}/{remote_name}".format(
                  endpoint=self.aws_endpoint_url,
                  remote_name=file_remote_name,
                  remote_path=self.remote_test_location.rstrip("/"),
                  local_name=file_local_name)
        ret = os.system(cmd)
        if ret != 0:
            logging.error("\033[1;31m[ERROR]\033[0m Failed while uploading object {name} with command {cmd}"
                  .format(name=file_local_name, cmd=cmd))
        return ret
        # self.s3_.meta.client.upload_file(file_local_name, 'ad_binary_release', remote_file)

    def release_object(self, object_name, commit_id=None, force=False):
        origin_path = self.find_proper_test_object(object_name, commit_id)
        dest_path = "s3://" + self.path_handler.RemoveDuplicateSlash(self.remote_release_location +"/" +object_name)
        if force:
            cmd = "aws --endpoint-url={endpoint} --profile ad_system_admin "\
                "s3 rm {dest_path}" \
                .format(endpoint=self.aws_endpoint_url, dest_path=dest_path)
            os.system(cmd)
        cmd = "aws --endpoint-url={endpoint} --profile ad_system_admin "\
              "s3 cp {origin_path} {dest_path}" \
              .format(endpoint=self.aws_endpoint_url, origin_path=origin_path, dest_path=dest_path)
        ret = os.system(cmd)
        if ret != 0:
            logging.error("\033[1;31m[ERROR]\033[0m Failed while release object from {origin_path} to {dest_path} with cmd: {cmd}"
                  .format(origin_path=origin_path, dest_path=dest_path, cmd=cmd))
            return ret
        return 0

if __name__ == "__main__":
    config_file = ""
    querier = Querier(config_file)
