#!/usr/bin/env python3

"""
provide Releaser for model management.
Releaser class is in charge of :
   synchronize local test objects
   move remote test objects to release
   print SHA256 code of objects that enable version releaser change manually the XXX_manifest.xml
"""

import sys
import os
import argparse
import logging
from hashlib import sha256
from .manifest_parser_and_generator import Object, Manifest
from .querier import Querier
from .synchronizer import Synchronizer,calculate_sha256
from .path_handler import PathHandler
from .config_reader import ConfigReader


class Releaser(object):

    def __init__(self, config_file):
        self.config_file = config_file
        self.querier = Querier(config_file)
        self.config = ConfigReader()
        self.config.read(config_file)
        self.path_handler = PathHandler()
        self.resource_path = self.path_handler.resource_path
        self.commitid = os.popen("git --git-dir={}/.git rev-parse --short HEAD"
                        .format(self.path_handler.ConfigurePath(self.config.commitid_reference_git))).readline().strip('\n')
        self.local_test_location = self.path_handler.ConfigurePath(self.config.modes["test"].local_location)

        self.manifest = Manifest()
        self.manifest_file = self.path_handler.ConfigurePath(self.config.manifest_file)
        if not os.path.exists(self.manifest_file):
            logging.error("Manifest file {} not found!".format(self.manifest_file))
            exit(1)
        self.manifest.parse_from_file(self.manifest_file)

    def release_all_object(self, force=False):
        if not self.manifest.is_valid():
            logging.error("\033[1;31m[ERROR]\033[0m Please check your manifest.")
            return False
        flag_success = True
        synchronizer = Synchronizer(self.config_file)
        synchronizer.sync()
        if synchronizer.is_synced():
            # step 1: remove all objects that are already released last time
            for target_objects in self.manifest.objects.values():
                for target_object in target_objects:
                    if target_object.mode == "release":
                        print("Removing all objects with name \033[1;33m{}\033[0m from aws test folder".format(target_object.name))
                        self.querier.remove_all_test_objects(target_object.name)
            # step 2:
            for target_objects in self.manifest.objects.values():
                for target_object in target_objects:
                    if target_object.mode == "test":
                        flag_success = not self.querier.release_object(target_object.name, commit_id=self.commitid, force=force) and flag_success
                        local_object_path = self.local_test_location + "/" + target_object.name
                        hashcode = calculate_sha256(local_object_path)
                        print("===========================================")
                        print("sha256 for object \033[1;33m{}\033[0m is \033[1;33m{}\033[0m"
                            .format(target_object.name, hashcode))
                        if flag_success:
                            self.manifest.change_to_release_and_add_sha256(target_object.name, hashcode)
                        # print("\033[1;33mplease change manually the manifest file, including the sha256 and the mode attribute.\033[0m")
        else:
            flag_success = False
        if flag_success:
            print("\033[1;32mGenerating new manifest.xml\033[0m")
            self.manifest.generate_to_file(self.manifest_file)
        if flag_success:
            print("\033[1;32m[{} BINARY RELEASE MANAGER] RELEASE SUCCESS\033[0m".format(self.config.project))
            print("DO NOT forget to commit the manifest file and push and merge")
        else:
            print("\033[1;31m[{} BINARY RELEASE MANAGER] RELEASE FAILED\033[0m".format(self.config.project), file=sys.stderr)

def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        '--force',
        required=False,
        dest='force',
        action="store_true",
        default=False,
        help='force the release, remove object if already exists'
    )
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    config_file = ""
    args = parse()
    releaser = Releaser(config_file)
    releaser.release_all_object(force=args.force)
