#!/usr/bin/env python3

"""
provide Synchronizer for object management.
Synchronizer class is in charge of :
   parse object config manifest
   check existance of local objects
   download missing objects from AWS
"""

import os
import sys
import logging
from hashlib import sha256
import argparse
from .manifest_parser_and_generator import Manifest, Object
from .path_handler import PathHandler
from .config_reader import ConfigReader
from .querier import Querier

def calculate_sha256(path_object):
    if os.path.isdir(path_object):
        # for a folder, use file size to calculate sha256 to accelerate
        files = []
        hashcodes = ["a"]
        # list all files in the dir
        contents = list()
        for root, dirs, files in os.walk(path_object):
            for f in files:
                contents.append(root+'/'+f)
        for file_path in contents:
            hashcode = sha256(str(os.stat(file_path).st_size).encode("utf8")).hexdigest()
            hashcodes.append(hashcode)
        hashcodes.sort()
        real_hash = sha256(str(hashcodes).encode("utf8")).hexdigest()
    else:
        # for a file, calculate directly the sha256
        with open(path_object,"rb") as f:
            bytes_read = f.read()
            real_hash = sha256(bytes_read).hexdigest()
    return real_hash



class Synchronizer(object):

    def Mkdir(self, dir_path):
        try:
            os.Mkdirs(dir_path)
        except:
            os.system("sudo mkdir -p {}".format(dir_path))

    def Remove(self, file_path):
        try:
            os.remove(file_path)
        except:
            os.system("sudo remove -rf {}".format(file_path))

    def __init__(self, config_file):
        self.config_file = config_file
        self.manifest = Manifest()
        self.is_synced_ = False
        path_handler = PathHandler()
        self.resource_path = path_handler.resource_path
        self.config = ConfigReader()
        self.config.read(config_file)
        self.default_manifest_file = path_handler.ConfigurePath(self.config.manifest_file)
        if os.path.exists(os.path.join(self.config.commitid_reference_git,".git")):
            self.commitid = os.popen("git --git-dir={}/.git rev-parse --short HEAD"
                            .format(path_handler.ConfigurePath(self.config.commitid_reference_git))).readline().strip('\n')
        else:
            self.commitid = None
        self.local_test_location = path_handler.ConfigurePath(self.config.modes["test"].local_location)
        if not os.path.exists(self.local_test_location):
            logging.warn("\033[1;33m[WARN]\033[0m {} does not exist. \n"
                    "       will mkdir, this may require root access.".format(self.local_test_location))
            ret = os.system("mkdir -p {} >/dev/null 2>&1".format(self.local_test_location))
            if 0 != ret:
                ret = os.system("sudo mkdir -p {}".format(self.local_test_location))
                if 0 != ret:
                    logging.error("\033[1;31m[ERROR]\033[0m Mkdir {} failed.".format(self.local_test_location))
                    sys.exit(1)

        self.local_release_location = path_handler.ConfigurePath(self.config.modes["release"].local_location)

        if not os.path.exists(self.local_release_location):
            logging.warn("\033[1;33m[WARN]\033[0m {} does not exist. \n"
                    "       will mkdir, this may require root access.".format(self.local_release_location))
            ret = os.system("mkdir -p {} >/dev/null 2>&1".format(self.local_release_location))
            if 0 != ret:
                ret = os.system("sudo mkdir -p {}".format(self.local_release_location))
                if 0 != ret:
                    logging.error("\033[1;31m[ERROR]\033[0m Mkdir {} failed.".format(self.local_release_location))
                    sys.exit(1)
        self.querier = None
        self.symbolic_link_path = path_handler.ConfigurePath(self.config.symbolic_dest)

    def check_existance(self, target_object):
        path_prefix = self.local_test_location * (target_object.mode == "test") \
                      + self.local_release_location * (target_object.mode == "release")
        path_full = path_prefix + "/" + target_object.name
        if os.path.exists(path_full):
            return True, path_full
        else:
            return False, path_full

    def download(self, target_object):
        if self.querier == None:
            self.querier = Querier(self.config_file)
        path_prefix = self.local_test_location * (target_object.mode == "test") \
                      + self.local_release_location * (target_object.mode == "release")
        logging.info("[INFO] Downloading model {name} to {path}..."
              .format(name=target_object.name,path=path_prefix))
        ret = self.querier.download_object(target_object.name, path_prefix, target_object.form, commit_id=self.commitid, object_mode=target_object.mode)
        if 0 != ret:
            return ret
        return ret

    def sync(self, manifest_file=None, selected_set=None):
        logging.info("[INFO] test object path is \033[0;33m{}\033[0m".format(self.local_test_location))
        logging.info("[INFO] release object path is \033[0;33m{}\033[0m".format(self.local_release_location))
        flag_success = True
        if not manifest_file or not os.path.exists(manifest_file):
            logging.warn("\033[1;33m[WARN]\033[0m Manifest path INVALID, will use default file.")
            manifest_file = self.default_manifest_file
        self.manifest.parse_from_file(manifest_file)
        if not self.manifest.is_valid():
            logging.error("\033[1;31m[ERROR]\033[0m Please check your manifest.")
            return False

        symlink_path = self.symbolic_link_path.rstrip('/')
        logging.info('[INFO] Check symbolic link path: {}'.format(symlink_path))
        if os.path.exists(symlink_path):
            if os.path.islink(symlink_path):
                logging.info('[INFO] {} is symbolic link, delete and create a directory'.format(symlink_path))
                self.Remove(symlink_path)
                self.Mkdir(symlink_path)

            elif not os.path.isdir(symlink_path):
                logging.info('[INFO] {} is not a directory, delete and create a directory'.format(symlink_path))
                self.Remove(symlink_path)
                self.Mkdir(symlink_path)

            else:
                logging.info('[INFO] {} already existed, will delete and create a new one'.format(symlink_path))
                ret = os.system("rm -rf {} 2>/dev/null".format(symlink_path))
                if not ret == 0:
                    os.system("sudo rm -rf {}".format(symlink_path))
                self.Mkdir(symlink_path)

        else:
            logging.info('[INFO] directory {} not found, try to create a new directory'.format(symlink_path))
            self.Mkdir(symlink_path)

        for target_objects in self.manifest.objects.values():
            for target_object in target_objects:
                flag_exist, path_object = self.check_existance(target_object)
                if not flag_exist:
                    logging.info("[INFO] Cannot find object {name} of mode {mode}, will download.".format(name=target_object.name,mode=target_object.mode))
                    flag_success = not self.download(target_object) and flag_success
                else:
                    logging.info("[INFO] Found object {name} of mode {mode}".format(name=target_object.name,mode=target_object.mode))
                    if target_object.mode == "release":
                        # Check SHA256
                        real_hash = calculate_sha256(path_object)
                        if target_object.sha256 != real_hash:
                            logging.error("\033[1;31m[ERROR]\033[0m SHA256 check for object {} failed.".format(target_object.name))
                            logging.error("desired sha256 is : \033[1;33m{desired}\033[0m\n  while the actual is : \033[1;33m{actual}\033[0m\n"
                                        "will remove and retry..."
                                .format(desired=target_object.sha256,actual=real_hash))
                            if os.path.isdir(path_object):
                                os.system("sudo rm -rf {}".format(path_object))
                            else:
                                os.system("sudo rm {}".format(path_object))
                            flag_success = not self.download(target_object) and flag_success
                # if selected_set provided, only create symlink for objects in selected set or objects with no set
                if selected_set and target_object.set and not (selected_set in target_object.set):
                    continue
                # Creating symbolic link
                if '/' in target_object.type:
                    symlink_dirname = os.path.dirname(os.path.join(self.symbolic_link_path, target_object.type))
                    if not os.path.exists(symlink_dirname):
                        try:
                            os.Mkdirs(symlink_dirname)
                        except:
                            os.system("sudo mkdir -p {}".format(symlink_dirname))
                symlink_full_path = os.path.join(self.symbolic_link_path, target_object.type)
                if os.path.exists(symlink_full_path):
                    try:
                        os.remove(symlink_full_path)
                    except:
                        os.system("sudo rm {}".format(symlink_full_path))
                try:
                    os.symlink(path_object, symlink_full_path)
                except:
                    os.system("sudo ln -sf {} {}".format(path_object, symlink_full_path))
        if flag_success:
            print("\033[1;32m[{} BINARY RELEASE MANAGER] SYNC SUCCESS\033[0m".format(self.config.project))
            self.is_synced_ = True
        else:
            print("\033[1;31m[{} BINARY RELEASE MANAGER] SYNC FAILED\033[0m".format(self.config.project), file=sys.stderr)

    def is_synced(self):
        return self.is_synced_

def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-m",
        '--manifest',
        required=False,
        dest='manifest',
        type=str,
        default=None,
        help='the model manifest file used to synchronize'
    )
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    config_file = ""
    args = parse()
    synchronizer = Synchronizer(config_file)
    synchronizer.sync(args.manifest)
    if synchronizer.is_synced_:
        sys.exit(0)
    else:
        sys.exit(1)
