#pragma once

#include <Eigen/Core>
#include "ad_common/data_type/base.hpp"
#include "ad_coordutils/data_type/status.hpp"

namespace saffron {

class CoordinateConverter {
 public:
    CoordinateConverter(CoordinateConverter const &) = delete;

    CoordinateConverter &operator=(CoordinateConverter const &) = delete;

    static CoordinateConverter *GetInstance() {
        static CoordinateConverter converter;
        return &converter;
    }

    void SetOrigin(const PointLLH_t &origin);

    coordStatus_t GetOrigin(PointLLH_t *origin) const;

    coordStatus_t LLA2ENU(const PointLLH_t &lla, PointENU_t *enu) const;

    coordStatus_t ENU2LLA(const PointENU_t &enu, PointLLH_t *lla) const;

    // Convert Geodetic coordinates to ECEF coordinates.
    void LLA2ECEF(const PointLLH_t &lla, PointENU_t *ecef) const;

    // Convert ECEF coordinates to Geodetic coordinates.
    void ECEF2LLA(const PointENU_t &ecef, PointLLH_t *lla) const;

 private:
    CoordinateConverter() = default;

    // Converts ECEF coordinate position into local-tangent-plane ENU.
    coordStatus_t ecef2Enu(const PointENU_t &ecef, PointENU_t *enu) const;

    // Converts ENU(East/North/Up) to ECEF coordinates
    coordStatus_t enu2Ecef(const PointENU_t &enu, PointENU_t *ecef) const;

    inline float64_t rad2Deg(float64_t radians) const {
        return (radians / M_PI) * 180.0;
    }

    inline float64_t deg2Rad(float64_t degrees) const {
        return (degrees / 180.0) * M_PI;
    }

    float64_t origin_longitude_rad_;
    float64_t origin_latitude_rad_;
    float64_t origin_altitude_m_;
    PointENU_t origin_ecef_m_;

    Eigen::Matrix3d matrix_ecef_to_enu_;
    Eigen::Matrix3d matrix_enu_to_ecef_;

    bool has_origin_ = false;

    // Geodetic system parameters
    //   https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
    static constexpr float64_t kSemimajorAxis = 6378137;
    static constexpr float64_t kSemiminorAxis = 6356752.3142;
    static constexpr float64_t kFirstEccentricitySquared =
        6.69437999014 * 0.001;
    static constexpr float64_t kSecondEccentricitySquared =
        6.73949674228 * 0.001;
    static constexpr float64_t kFlattening = 1 / 298.257223563;
};

}  // namespace saffron
