#pragma once

#include <vector>
#include <string>
#include <map>

#include "ad_common/data_type/base.hpp"
#include "map/common/error.hpp"

namespace saffron {
namespace hdm {
enum class LaneType {
    NONE = 0,
    DRIVING,
    LEFT_WAITING,
    RIGHT_WAITING,
    STRAIGHT_WAITING,
    BUS,
    BICYCLE,
    EMERGENCY
};
using LaneTypeTable_t = std::map<LaneType, std::string>;
static const LaneTypeTable_t LaneTypeTable{
    {LaneType::NONE, "NONE"},
    {LaneType::DRIVING, "driving"},
    {LaneType::LEFT_WAITING, "left_waiting"},
    {LaneType::RIGHT_WAITING, "right_waiting"},
    {LaneType::STRAIGHT_WAITING, "straight_waiting"},
    {LaneType::BUS, "bus"},
    {LaneType::BICYCLE, "bicycle"},
    {LaneType::EMERGENCY, "emergency"}};

enum class RoadType { NONE = 0, NORMAL, WAITING, SIDE, ROUNDABOUT };
using RoadTypeTable_t = std::map<RoadType, std::string>;
static const RoadTypeTable_t RoadTypeTable{
    {RoadType::NONE, "NONE"},
    {RoadType::NORMAL, "normal"},
    {RoadType::WAITING, "waiting"},
    {RoadType::SIDE, "side"},
    {RoadType::ROUNDABOUT, "roundabout"}};

enum class TurnType {
    FREE = 0xffff,
    TURN_LEFT = 0x0001,     // 0000 0001 determine turn left
    TURN_RIGHT = 0x0002,    // 0000 0010  determine turn right
    LEFT_AROUND = 0x0004,   // 0000 0100 determine turn left around
    RIGHT_AROUND = 0x0008,  // 0000 1000 determine turn right around
    STRAIGHT = 0x0010       // 0001 0000 determine go straight
};

enum class TurnLink {
    NONE = 0,
    LEFT,
    RIGHT,
    LEFT_AROUND,
    RIGHT_AROUND,
    STRAIGHT
};
using TurnLinkTable_t = std::map<TurnLink, std::string>;
static const TurnLinkTable_t TurnLinkTable{
    {TurnLink::NONE, "NONE"},
    {TurnLink::LEFT, "left"},
    {TurnLink::RIGHT, "right"},
    {TurnLink::LEFT_AROUND, "left_around"},
    {TurnLink::RIGHT_AROUND, "right_around"},
    {TurnLink::STRAIGHT, "straight"}};

enum class LineType {
    NONE = 0x0000,
    WHITE = 0x0001,     // 0 bit determine white
    YELLOW = 0x0002,    // 1 bit determine yellow
    SOLID = 0x0004,     // 2 bit determine solid or dash
    ROADSIDE = 0x0008,  // 3 bit determine roadside
    VIRTUAL = 0x0010    // 4 bit determine go straight
};

struct Header {
    std::string version;
    std::string date;
    std::string vendor;
    bool right_hand_drive;
    PointLLH_t origin;
};
struct Boundary {
    bool is_virtual;
    std::vector<Point3D_t> geometry;
};

struct TrafficLightData {
    std::string id;  // traffic_light_xx
    std::vector<Point3D_t> geometry;
    // left_arrow, right_arrow, circle, left_around, right_around
    std::string type;
};

struct GroupTrafficLightData {
    std::string id;                              // group_traffic_light_xx
    std::vector<std::string> traffic_light_ids;  // traffic lights in same group
    uint16_t turn_type;
};

struct ObjectData {
    std::string id;  // obstacle_xx, sidewalk_xx, stop_line_xx, pole_xx,
                     // road_marker_xx, park_area_xx
    std::vector<Point3D_t> geometry;
};

struct CrosswalkData {
    std::string id;  // crosswalk_xx
    std::vector<Point3D_t> geometry;
    std::vector<std::string> contain_ids;
};

struct GuideSignData {
    std::string id;  // guide_sign_xx
    std::vector<Point3D_t> geometry;
    std::string type;  // circle, rectangle
};

struct JunctionData {
    std::string id;  // junction_xx
    std::vector<std::string> overlap_ids;
    std::vector<Boundary> boundaries;
};

struct RoadData {
    std::string id;  // road_xx
    std::string name;
    float64_t length;
    std::vector<std::string> overlap_ids;
    std::vector<std::string> section_ids;
    std::vector<std::string> predecessor_ids;
    std::vector<std::string> successor_ids;
    RoadType type;
};

struct SectionData {
    std::string id;  // section_xx
    std::string road_id;
    float64_t length;
    std::vector<std::string> lane_ids;  // from left to right
    std::vector<std::string> overlap_ids;
    std::string left_boundary_id;
    std::string right_boundary_id;
    std::vector<std::string> predecessor_ids;
    std::vector<std::string> successor_ids;
};

struct NeighborLane {
    std::string lane_id;
    bool forward;
    bool change;
};

struct LaneLinkData {
    std::string id;  // lane_link_xx
    std::string from_lane_id;
    std::string to_lane_id;
    std::string ref_line_id;
    TurnLink turn_link;
};

struct RoadLinkData {
    std::string id;  // road_link_xx
    std::string from_road_id;
    std::string to_road_id;
    TurnLink turn_link;
};

struct SectionLinkData {
    std::string id;  // section_link_xx
    std::string from_section_id;
    std::string to_section_id;
    TurnLink turn_link;
};

struct LaneData {
    std::string id;  // lane_xx
    std::string road_id;
    std::string section_id;
    std::string left_line_id;
    std::string right_line_id;
    std::string center_line_id;
    std::vector<std::string> overlap_ids;
    LaneType lane_type;
    float64_t max_speed;
    float64_t min_speed;
    uint16_t turn_type;
    float64_t length;
    std::vector<float64_t> widths;  // size = center_line.size
    NeighborLane left_neighbor_lane;
    NeighborLane right_neighbor_lane;
    std::vector<std::string> successor_ids;
    std::vector<std::string> predecessor_ids;
};

struct LineData {
    std::string id;  // line_xx
    std::vector<Point3D_t> geometry;
    float64_t width;
    uint16_t line_type;
};
struct RoutingMapData {
    std::vector<LaneData> lanes;
    std::vector<JunctionData> junctions;
    std::vector<RoadData> roads;
    std::vector<SectionData> sections;
    std::vector<LaneLinkData> lane_links;
    std::vector<RoadLinkData> road_links;
    std::vector<SectionLinkData> section_links;
    std::vector<GroupTrafficLightData> group_traffic_lights;
};
struct SemanticMapData {
    std::vector<LineData> lines;
    std::vector<GuideSignData> guide_signs;
    std::vector<TrafficLightData> traffic_lights;
    std::vector<ObjectData> obstacles;
    std::vector<ObjectData> sidewalks;
    std::vector<ObjectData> stop_lines;
    std::vector<ObjectData> poles;
    std::vector<ObjectData> road_markers;
    std::vector<CrosswalkData> crosswalks;
    std::vector<ObjectData> park_areas;
};

struct HDMapData {
    Header header;
    RoutingMapData routing_map_data;
    SemanticMapData semantic_map_data;
};
}  // namespace hdm
}  // namespace saffron
