#pragma once
#include <string>
#include <memory>
#include <vector>
#include <mutex>
#include <opencv2/core.hpp>

#include "map/data_type/hdmap_data.hpp"
#include "map/serialization/hdmap_serialize.hpp"
#include "map/hdmap_param.hpp"
#include "map/routing_map/lane.hpp"
#include "map/routing_map/section.hpp"
#include "map/routing_map/road.hpp"
#include "map/routing_map/junction.hpp"
#include "map/routing_map/group_traffic_light.hpp"
#include "map/semantic_map/line.hpp"
#include "map/semantic_map/object.hpp"
#include "map/data_type/routing_data.hpp"

namespace saffron {
namespace hdm {
class HDMapImpl;
class RoutingImpl;

class HDMap {
 public:
    HDMap();
    ~HDMap();
    /**
     * @brief Load map from file
     */
    mapStatus_t LoadMap(const std::string& file_name);
    /**
     * @brief Load map origin from file, in case of some node just want to get
     * origin point.
     */
    mapStatus_t LoadMapOrigin(const std::string& file_name);

    /**
     * @brief Load map origin from origin data
     */
    mapStatus_t LoadMap(HDMapData&& hdmap_data);
    /**
     * @brief Load map origin from origin data
     */
    mapStatus_t LoadMap(const HDMapData& hdmap_data);
    /**
     * @brief init param
     */
    mapStatus_t Init(const HDMapParam& param);

    /**
     * @brief Get map header
     */
    const Header& GetHeader();

    /**
     * @brief Get map param
     */
    const HDMapParam& GetParam();

    /**
     * @brief Get range map data of specified distance
     */
    mapStatus_t GetRangeHDMapData(const Point3D_t& position,
                                  float64_t distance,
                                  HDMapData* hdmap_data);
    /**
     * @brief Get range map data of param.query_radius, default is 300m
     */
    mapStatus_t GetRangeHDMapData(const Point3D_t& position,
                                  HDMapData* hdmap_data);
    /************************occupied map**************************/
    /************************occupied map**************************/
    /************************occupied map**************************/
    /**
     * @brief Gompute freespace of param.query_radius, default is 300m
     */
    mapStatus_t ComputeFreeSpace(const Point3D_t& position, float64_t heading);
    /**
     * @brief Get freespace
     */
    mapStatus_t GetFreeSpace(
        std::vector<std::vector<cv::Point2f>>* freespace_contours,
        std::vector<cv::Vec4i>* freespace_hierarchy);
    mapStatus_t GetVehicleCoorFreeSpace(
        const Point3D_t& position,
        float64_t heading,
        std::vector<std::vector<cv::Point2f>>* freespace_contours,
        std::vector<cv::Vec4i>* freespace_hierarchy);
    /************************routing map**************************/
    /************************routing map**************************/
    /************************routing map**************************/
    mapStatus_t GetAllSection(Section::Vector* all_section);
    mapStatus_t GetAllLane(Lane::Vector* all_lane);
    mapStatus_t GetAllJunction(Junction::Vector* all_junction);
    mapStatus_t GetAllRoad(Road::Vector* all_road);
    mapStatus_t GetAllLaneLink(LaneLink::Vector* all_lane_link);
    mapStatus_t GetAllSectionLink(SectionLink::Vector* all_section_link);
    mapStatus_t GetAllRoadLink(RoadLink::Vector* all_road_link);
    mapStatus_t GetAllGroupTrafficLight(GroupTrafficLight::Vector* all_gtl);
    bool IsRightHandDrive();
    /**
     * @brief Get element by id, return null when can't find
     */
    Lane::ConstPtr GetLaneById(const std::string& id);
    Section::ConstPtr GetSectionById(const std::string& id);
    Road::ConstPtr GetRoadById(const std::string& id);
    Junction::ConstPtr GetJunctionById(const std::string& id);
    LaneLink::ConstPtr GetLaneLinkById(const std::string& id);
    GroupTrafficLight::ConstPtr GetGroupTrafficLightById(const std::string& id);
    SectionLink::ConstPtr GetSectionLinkById(const std::string& id);
    RoadLink::ConstPtr GetRoadLinkById(const std::string& id);

    /**
     * @brief Get roads to which sections belong.
     */
    mapStatus_t GetRoadsBySections(const Section::Vector& secs,
                                   Road::Vector* roads);
    /**
     * @brief Get sections contained in sections
     */
    mapStatus_t GetLanesBySections(const Section::Vector& secs,
                                   Lane::Vector* lanes);
    mapStatus_t GetLanesBySection(const Section::ConstPtr& sec,
                                  Lane::Vector* lanes);

    /**
     * @brief Get lane links contained in lanes
     */
    mapStatus_t GetLaneLinkByLanes(const Lane::Vector& lanes,
                                   LaneLink::Vector* lane_links);

    /**
     * @brief Get link connecting two lanes/sections/roads
     */
    LaneLink::ConstPtr GetLaneLinkBetweenLanes(const Lane::ConstPtr& from_lane,
                                               const Lane::ConstPtr& to_lane);
    SectionLink::ConstPtr GetSectionLinkBetweenSections(
        const Section::ConstPtr& from_section,
        const Section::ConstPtr& to_section);
    RoadLink::ConstPtr GetRoadLinkBetweenRoads(const Road::ConstPtr& from_road,
                                               const Road::ConstPtr& to_road);
    /**
     * @brief Get range sections/junctions/lanes
     */
    mapStatus_t GetRangeSections(const PointENU_t& position,
                                 double distance,
                                 Section::Vector* sections);
    mapStatus_t GetRangeJunctions(const PointENU_t& position,
                                  double distance,
                                  Junction::Vector* junctions);
    mapStatus_t GetRangeLanes(const PointENU_t& position,
                              double distance,
                              Lane::Vector* lanes);

    /**
     * @brief If vehicle's heading == lane's heading. Because lane  may be
     * curved, please make sure input lane and position is near.
     */
    bool IsAlongTheLane(const Point3D_t& position,
                        float64_t heading,
                        const Lane::ConstPtr& lane);

    /**
     * @brief get straight drivable lane rank of section(from left to right);
     */
    size_t GetStraightLaneRankOfSection(const Lane::ConstPtr& cur_lane);
    size_t GetDrivableLaneRankOfSection(const Lane::ConstPtr& cur_lane);
    /**
     * @brief get straight drivable lanes  of section(from left to right);
     */
    mapStatus_t GetStraightLanesOfSection(const Section::ConstPtr& cur_section,
                                          Lane::Vector* lanes);

    mapStatus_t GetDrivableLanesOfSection(const Section::ConstPtr& cur_section,
                                          Lane::Vector* lanes);
    /**
     * @brief If input lane_link is into waiting zone.
     */
    bool IsIntoWaiting(const LaneLink::ConstPtr& lane_link);

    /**
     * @brief Get current junction, only position is literally outside junction.
     */
    Junction::ConstPtr GetCurrentJunction(const Point3D_t& position);

    /**
     * @brief Get forward nearest junction. vehicle must be on the lane.
     */
    Junction::ConstPtr GetForwardNearestJunction(
        const Point3D_t& position,
        float64_t heading,
        const Lane::ConstPtr& lane_ptr = nullptr);

    /**
     * @brief Get backward nearest junction. vehicle must be on the lane.
     */
    Junction::ConstPtr GetBackwardNearestJunction(
        const Point3D_t& position,
        float64_t heading,
        const Lane::ConstPtr& lane_ptr = nullptr);
    /**
     * @brief Get literally nearest junction.
     */
    Junction::ConstPtr GetNearestJunction(const Point3D_t& position);

    /**
     * @brief Get current lane, if you already know current section, can pass it
     * for speed up. return nullptr when position is literally outside lane.
     */
    Lane::ConstPtr GetCurrentLane(const Point3D_t& position,
                                  Section::ConstPtr section = nullptr);
    /**
     * @brief Get literally nearest lane, if you already know current section,
     * can pass it for speed up.
     */
    Lane::ConstPtr GetNearestLane(const Point3D_t& position,
                                  Section::ConstPtr section = nullptr);

    /**
     * @brief Get current section, return nullptr when position is literally
     * outside section.
     */
    Section::ConstPtr GetCurrentSection(const Point3D_t& position);
    /**
     * @brief Get literally nearest section.
     */
    Section::ConstPtr GetNearestSection(const Point3D_t& position);

    /**
     * @brief Get current road, return nullptr when position is literally
     * outside road.
     */
    Road::ConstPtr GetCurrentRoad(const Point3D_t& position);
    /**
     * @brief Get current road, return nullptr when position is literally
     * outside road.
     */
    Road::ConstPtr GetNearestRoad(const Point3D_t& position);

    /**
     * @brief Get Successor links of input lane.
     */
    mapStatus_t GetSuccessorLinks(const Lane::ConstPtr& lane,
                                  LaneLink::Vector* lane_links);
    mapStatus_t GetSuccessorLinks(const Section::ConstPtr& section,
                                  SectionLink::Vector* section_links);
    mapStatus_t GetSuccessorLinks(const Road::ConstPtr& road,
                                  RoadLink::Vector* road_links);

    /**
     * @brief Get Successor lanes of input lane.
     */
    mapStatus_t GetSuccessors(const Lane::ConstPtr& lane, Lane::Vector* lanes);
    mapStatus_t GetSuccessors(const Section::ConstPtr& section,
                              Section::Vector* sections);
    mapStatus_t GetSuccessors(const Road::ConstPtr& road, Road::Vector* roads);

    /**
     * @brief Get Successor lane of input lane_link.
     */
    Lane::ConstPtr GetSuccessor(const LaneLink::ConstPtr& lane_link);
    Section::ConstPtr GetSuccessor(const SectionLink::ConstPtr& section_link);
    Road::ConstPtr GetSuccessor(const RoadLink::ConstPtr& road_link);

    /**
     * @brief Get predecesor links of input lane.
     */
    mapStatus_t GetPredecessorLinks(const Lane::ConstPtr& lane,
                                    LaneLink::Vector* lane_links);
    mapStatus_t GetPredecessorLinks(const Section::ConstPtr& section,
                                    SectionLink::Vector* section_links);
    mapStatus_t GetPredecessorLinks(const Road::ConstPtr& road,
                                    RoadLink::Vector* road_links);
    /**
     * @brief Get predecesor lanes of input lane.
     */
    mapStatus_t GetPredecessors(const Lane::ConstPtr& lane,
                                Lane::Vector* lanes);
    mapStatus_t GetPredecessors(const Section::ConstPtr& section,
                                Section::Vector* sections);
    mapStatus_t GetPredecessors(const Road::ConstPtr& road,
                                Road::Vector* roads);
    /**
     * @brief Get predecesor lane of input lane_link.
     */
    Lane::ConstPtr GetPredecessor(const LaneLink::ConstPtr& lane_link);
    Section::ConstPtr GetPredecessor(const SectionLink::ConstPtr& section_link);
    Road::ConstPtr GetPredecessor(const RoadLink::ConstPtr& road_link);

    /**
     * @brief Get left/right neighbor lane of input lane.
     */
    Lane::ConstPtr GetLeftNeighborLane(const Lane::ConstPtr& current_lane);
    Lane::ConstPtr GetRightNeighborLane(const Lane::ConstPtr& current_lane);

    /**
     * @brief Get left/right/center lane line of input lane.
     */
    Line::ConstPtr GetLeftLaneLine(const Lane::ConstPtr& lane);
    mapStatus_t GetLeftLaneLineGeo(const Lane::ConstPtr& lane,
                                   std::vector<Point3D_t>* geo);
    Line::ConstPtr GetRightLaneLine(const Lane::ConstPtr& lane);

    mapStatus_t GetRightLaneLineGeo(const Lane::ConstPtr& lane,
                                    std::vector<Point3D_t>* geo);
    Line::ConstPtr GetCenterLaneLine(const Lane::ConstPtr& lane);
    mapStatus_t GetCenterLaneLineGeo(const Lane::ConstPtr& lane,
                                     std::vector<Point3D_t>* geo);

    /**
     * @brief Get left/right bound line of input section.
     */
    Line::ConstPtr GetLeftSectionBound(const Section::ConstPtr& sec);
    Line::ConstPtr GetRightSectionBound(const Section::ConstPtr& sec);

    /**
     * @brief Get remain forward section ids in one road of vehicle's position.
     */
    mapStatus_t GetForwardSectionIds(
        const Point3D_t& position,
        std::vector<std::string>* sec_ids,
        const Lane::ConstPtr& current_lane = nullptr);
    /**
     * @brief Get remain backward section ids in one road of vehicle's position.
     */
    mapStatus_t GetBackwardSectionIds(
        const Point3D_t& position,
        std::vector<std::string>* sec_ids,
        const Lane::ConstPtr& current_lane = nullptr);

    /**
     * @brief Get current lane width. return -1 when vehicle doesn't in any
     * lane.
     */
    float64_t GetCurrentLaneWidth(const Point3D_t& position,
                                  Lane::ConstPtr current_lane = nullptr);
    /**
     * @brief Distance to left/right lane line. return max when vehicle doesn't
     * in any lane.
     */
    float64_t Dis2LeftLaneLine(const Point3D_t& position,
                               Lane::ConstPtr current_lane = nullptr);
    float64_t Dis2RightLaneLine(const Point3D_t& position,
                                Lane::ConstPtr current_lane = nullptr);
    /**
     * @brief Distance to lane/road begin/end. return max when vehicle doesn't
     * in any lane or road is longer than param.query_radius.
     */
    float64_t Dis2LaneEnd(const Point3D_t& position,
                          Lane::ConstPtr current_lane = nullptr);
    float64_t Dis2LaneBegin(const Point3D_t& position,
                            Lane::ConstPtr current_lane = nullptr);
    float64_t Dis2RoadEnd(const Point3D_t& position,
                          const Lane::ConstPtr& current_lane = nullptr);
    float64_t Dis2RoadBegin(const Point3D_t& position,
                            Lane::ConstPtr current_lane = nullptr);

    /**
     * @brief Distance to forward/backward stop_line/junction. return max when
     * vehicle doesn't in any lane or dis to junction/stop_line is longer than
     * param.query_radius.
     */
    float64_t Dis2ForwardStopLine(const Point3D_t& position,
                                  float64_t heading,
                                  const Lane::ConstPtr& current_lane = nullptr);
    float64_t Dis2ForwardJunction(const Point3D_t& position,
                                  float64_t heading,
                                  const Lane::ConstPtr& current_lane = nullptr);
    float64_t Dis2BackwardJunction(
        const Point3D_t& position,
        float64_t heading,
        const Lane::ConstPtr& current_lane = nullptr);

    /**
     * @brief Get go_straight/turn_left/etc lanes of input current lane. The
     * returned lanes is empty when input lane is not overlapped junction,
     * except straight. Because lane_link's turn link is straight by default.
     */
    mapStatus_t GetGoStraightLanes(const Lane::ConstPtr& lane,
                                   Lane::Vector* lanes);
    mapStatus_t GetTurnLeftLanes(const Lane::ConstPtr& lane,
                                 Lane::Vector* lanes);
    mapStatus_t GetTurnRightLanes(const Lane::ConstPtr& lane,
                                  Lane::Vector* lanes);
    mapStatus_t GetTurnLeftAroundLanes(const Lane::ConstPtr& lane,
                                       Lane::Vector* lanes);
    mapStatus_t GetTurnRightAroundLanes(const Lane::ConstPtr& lane,
                                        Lane::Vector* lanes);
    /**
     * @brief Query Nearest reflines when vehicle at junction.
     */
    mapStatus_t QueryNearestRefLines(const Point3D_t& position,
                                     float64_t heading,
                                     LaneLink::Vector* lane_links,
                                     Junction::ConstPtr junction_ptr = nullptr,
                                     float64_t radian_threshold = M_PI_4,
                                     float64_t distance_threshold = 8.0,
                                     size_t num_threshold = 5);

    /**
     * @brief Get lane_links which are get through a junction.
     */
    mapStatus_t GetLaneLinkOfJunction(const Junction::ConstPtr& junction_ptr,
                                      LaneLink::Vector* lane_links);

    mapStatus_t GetLaneLinesBySections(const Section::Vector& sections,
                                       Line::Vector* lines);

    /**
     * @brief Get center point
     */
    mapStatus_t GetCenterPoint(const Section::ConstPtr& section, Point3D_t* pt);
    mapStatus_t GetCenterPoint(const Lane::ConstPtr& lane, Point3D_t* pt);
    mapStatus_t GetCenterPoint(const Junction::ConstPtr& junction,
                               Point3D_t* pt);
    /**
     * @brief Get ref_lines by lane_links/.
     */
    mapStatus_t GetRefLinesByLaneLinks(const LaneLink::Vector& lane_links,
                                       Line::Vector* lines);
    mapStatus_t GetRefLinesByPredecessor(const Lane::ConstPtr& lane_ptr,
                                         Line::Vector* lines);
    mapStatus_t GetRefLinesBySuccessor(const Lane::ConstPtr& lane_ptr,
                                       Line::Vector* lines);
    /**
     * @brief Get ref_line by lane_link.
     */
    Line::ConstPtr GetRefLine(const LaneLink::ConstPtr& lane_link);

    /************************semantic map**************************/
    /************************semantic map**************************/
    /************************semantic map**************************/

    /**
     * @brief Get current cross walk, return nullptr when position is literally
     * outside cross walk.
     */
    Crosswalk::ConstPtr GetCurrentCrosswalk(const Point3D_t& position);

    /**
     * @brief Get All element.
     */
    mapStatus_t GetAllLine(Line::Vector* all_line);
    mapStatus_t GetAllGuideSign(GuideSign::Vector* all_guide_sign);
    mapStatus_t GetAllTrafficLight(TrafficLight::Vector* all_traffic_light);
    mapStatus_t GetAllObstacle(Obstacle::Vector* all_obstacle);
    mapStatus_t GetAllStopLine(StopLine::Vector* all_stop_line);
    mapStatus_t GetAllRoadMarker(RoadMarker::Vector* all_road_marker);
    mapStatus_t GetAllCrosswalk(Crosswalk::Vector* all_crosswalk);
    mapStatus_t GetAllParkArea(ParkArea::Vector* all_park_area);
    mapStatus_t GetAllSidewalk(Sidewalk::Vector* all_sidewalk);
    mapStatus_t GetAllPole(Pole::Vector* all_pole);

    /**
     * @brief Get forward nearest stop line, vehcle must be in a lane.
     */
    StopLine::ConstPtr GetForwardNearestStopLine(
        const Point3D_t& position,
        float64_t heading,
        const Lane::ConstPtr& lane = nullptr);
    /**
     * @brief Get forward group traffic lights, vehcle must be in a lane.
     */
    mapStatus_t GetForwardGroupTrafficLights(
        const Point3D_t& position,
        GroupTrafficLight::Vector* group_tls,
        const Lane::ConstPtr& lane = nullptr);

    /**
     * @brief Get element by id. Return nullptr when can't find it.
     */
    Line::ConstPtr GetLineById(const std::string& id);
    GuideSign::ConstPtr GetGuideSignById(const std::string& id);
    TrafficLight::ConstPtr GetTrafficLightById(const std::string& id);
    Obstacle::ConstPtr GetObstacleById(const std::string& id);
    StopLine::ConstPtr GetStopLineById(const std::string& id);
    RoadMarker::ConstPtr GetRoadMarkerById(const std::string& id);
    Crosswalk::ConstPtr GetCrosswalkById(const std::string& id);
    ParkArea::ConstPtr GetParkAreaById(const std::string& id);
    Sidewalk::ConstPtr GetSideWalkById(const std::string& id);
    Pole::ConstPtr GetPoleById(const std::string& id);

    mapStatus_t GetLinesBySections(const Section::Vector& sections,
                                   Line::Vector* lines);

    /**
     * @brief Get range elements of distance
     */
    mapStatus_t GetRangeLines(const PointENU_t& position,
                              double distance,
                              Line::Vector* lines);
    mapStatus_t GetRangeTrafficLights(const PointENU_t& position,
                                      double distance,
                                      TrafficLight::Vector* traffic_lights);
    mapStatus_t GetRangeStopLines(const PointENU_t& position,
                                  double distance,
                                  StopLine::Vector* stop_lines);
    mapStatus_t GetRangeSidewalks(const PointENU_t& position,
                                  double distance,
                                  Sidewalk::Vector* sidewalks);
    mapStatus_t GetRangePoles(const PointENU_t& position,
                              double distance,
                              Pole::Vector* poles);
    mapStatus_t GetRangeRoadMarkers(const PointENU_t& position,
                                    double distance,
                                    RoadMarker::Vector* road_markers);
    mapStatus_t GetRangeCrosswalks(const PointENU_t& position,
                                   double distance,
                                   Crosswalk::Vector* crosswalks);
    mapStatus_t GetRangeParkAreas(const PointENU_t& position,
                                  double distance,
                                  ParkArea::Vector* park_areas);
    mapStatus_t GetRangeGuideSigns(const PointENU_t& position,
                                   double distance,
                                   GuideSign::Vector* guide_signs);
    mapStatus_t GetRangeObstacles(const PointENU_t& position,
                                  double distance,
                                  Obstacle::Vector* obstacles);

    /************************ routing **************************/
    /************************ routing **************************/
    /************************ routing **************************/
    mapStatus_t InitRoutingGraph();

    mapStatus_t FindRoute(const std::vector<PointLLH_t>& way_points,
                         routing::RoutingResponse* routing_response,
                         bool pullover = true,
                         bool enable_u_turn = false,
                         const std::vector<PointLLH_t>& black_list = {});

    mapStatus_t FindRoute(const std::vector<PointENU_t>& way_points,
                         routing::RoutingResponse* routing_response,
                         bool pullover = true,
                         bool enable_u_turn = false,
                         const std::vector<PointENU_t>& black_list = {});

    mapStatus_t FindRoadRoute(const std::vector<PointLLH_t>& way_points,
                             std::vector<std::string>* road_list,
                             bool enable_u_turn = false);

    mapStatus_t LoadRoute(const std::string& path_file,
                         routing::RoutingResponse* routing_response,
                         bool online_routing = false,
                         bool enable_u_turn = false);

    mapStatus_t SaveRoute(const std::string& data_dir,
                         const routing::RoutingResponse& routing_response);

    mapStatus_t ToNavigationLine(
        const std::vector<routing::RoutingNode>& node_list,
        const PointLLH_t& start,
        const PointLLH_t& destination,
        std::vector<PointLLH_t>* navi_line);
    size_t FindIndexInRoutingResponse(const PointENU_t& position,
                                  const routing::RoutingResponse& routing_response);

 private:
    std::mutex hdmap_mutex_;
    std::unique_ptr<HDMapImpl> impl_;
    std::unique_ptr<RoutingImpl> routing_impl_;
};

class HDMapUtil {
 public:
    HDMapUtil(HDMapUtil const&) = delete;
    HDMapUtil& operator=(HDMapUtil const&) = delete;

    static HDMap* GetInstance() {
        static HDMap hdmap_;
        return &hdmap_;
    }

 private:
    HDMapUtil() = default;
};
}  // namespace hdm
}  // namespace saffron
