#pragma once
#include <vector>
#include <memory>
#include <unordered_map>
#include <algorithm>
#include <limits>
#include <string>
#include <Eigen/Core>
#include "map/math_tools/math_util.hpp"

namespace saffron {
namespace hdm {

struct BoxKDTreeParams {
    /// The maximum depth of the kdtree.
    int max_depth = -1;
    /// The maximum number of items in one leaf node.
    int max_leaf_size = -1;
    /// The maximum dimension size of leaf node.
    float64_t max_leaf_dimension = -1.0;
};

template <class Object>
class BoxKDTree2dNode {
 public:
    using ObjectPtr = std::shared_ptr<Object>;
    using ObjectConstPtr = std::shared_ptr<const Object>;
    /**
     * @brief Constructor which takes a vector of objects,
     *        parameters and depth of the node.
     * @param objects Objects to build the KD-tree node.
     * @param params Parameters to build the KD-tree.
     * @param depth Depth of the KD-tree node.
     */
    BoxKDTree2dNode(const std::vector<ObjectConstPtr> &objects,
                    const BoxKDTreeParams &params,
                    int depth)
        : depth_(depth) {
        assert(!objects.empty());

        ComputeBoundary(objects);
        ComputePartition();

        if (SplitToSubNodes(objects, params)) {
            std::vector<ObjectConstPtr> left_subnode_objects;
            std::vector<ObjectConstPtr> right_subnode_objects;
            PartitionObjects(objects, &left_subnode_objects,
                             &right_subnode_objects);

            // Split to sub-nodes.
            if (!left_subnode_objects.empty()) {
                left_subnode_.reset(new BoxKDTree2dNode<Object>(
                    left_subnode_objects, params, depth + 1));
            }
            if (!right_subnode_objects.empty()) {
                right_subnode_.reset(new BoxKDTree2dNode<Object>(
                    right_subnode_objects, params, depth + 1));
            }
        } else {
            InitObjects(objects);
        }
    }

    /**
     * @brief Get the nearest object to a target point by the KD-tree
     *        rooted at this node.
     * @param point The target point. Search it's nearest object.
     * @return The nearest object to the target point.
     */
    void GetNearestObject(const Eigen::Vector2d &point,
                          ObjectConstPtr *object,
                          float64_t *dis) const {
        GetNearestObjectInternal(point, object, dis);
    }

    /**
     * @brief Get objects within a distance to a point by the KD-tree
     *        rooted at this node.
     * @param point The center point of the range to search objects.
     * @param distance The radius of the range to search objects.
     * @return All objects within the specified distance to the specified
     point.
     */
    void GetObjects(const Eigen::Vector2d &point,
                    float64_t distance,
                    std::vector<ObjectConstPtr> *objects) const {
        GetObjectsInternal(point, distance, objects);
    }

 private:
    void InitObjects(const std::vector<ObjectConstPtr> &objects) {
        num_objects_ = static_cast<int>(objects.size());
        objects_sorted_by_min_ = objects;
        objects_sorted_by_max_ = objects;
        std::sort(objects_sorted_by_min_.begin(), objects_sorted_by_min_.end(),
                  [&](ObjectConstPtr obj1, ObjectConstPtr obj2) {
                      return partition_ == PARTITION_X
                                 ? obj1->GetPoly2d().MinX() <
                                       obj2->GetPoly2d().MinX()
                                 : obj1->GetPoly2d().MinY() <
                                       obj2->GetPoly2d().MinY();
                  });
        std::sort(objects_sorted_by_max_.begin(), objects_sorted_by_max_.end(),
                  [&](ObjectConstPtr obj1, ObjectConstPtr obj2) {
                      return partition_ == PARTITION_X
                                 ? obj1->GetPoly2d().MaxX() >
                                       obj2->GetPoly2d().MaxX()
                                 : obj1->GetPoly2d().MaxY() >
                                       obj2->GetPoly2d().MaxY();
                  });
        objects_sorted_by_min_bound_.reserve(num_objects_);
        for (ObjectConstPtr object : objects_sorted_by_min_) {
            objects_sorted_by_min_bound_.push_back(
                partition_ == PARTITION_X ? object->GetPoly2d().MinX()
                                          : object->GetPoly2d().MinY());
        }
        objects_sorted_by_max_bound_.reserve(num_objects_);
        for (ObjectConstPtr object : objects_sorted_by_max_) {
            objects_sorted_by_max_bound_.push_back(
                partition_ == PARTITION_X ? object->GetPoly2d().MaxX()
                                          : object->GetPoly2d().MaxY());
        }
    }

    bool SplitToSubNodes(const std::vector<ObjectConstPtr> &objects,
                         const BoxKDTreeParams &params) {
        if (params.max_depth >= 0 && depth_ >= params.max_depth) {
            return false;
        }
        if (static_cast<int>(objects.size()) <=
            std::max(1, params.max_leaf_size)) {
            return false;
        }
        if (params.max_leaf_dimension >= 0.0 &&
            std::max(max_x_ - min_x_, max_y_ - min_y_) <=
                params.max_leaf_dimension) {
            return false;
        }
        return true;
    }

    float64_t LowerDistanceToPoint(const Eigen::Vector2d &point) const {
        float64_t dx = 0.0;
        if (point.x() < min_x_) {
            dx = min_x_ - point.x();
        } else if (point.x() > max_x_) {
            dx = point.x() - max_x_;
        }
        float64_t dy = 0.0;
        if (point.y() < min_y_) {
            dy = min_y_ - point.y();
        } else if (point.y() > max_y_) {
            dy = point.y() - max_y_;
        }
        return std::hypot(dx, dy);
    }

    float64_t UpperDistanceToPoint(const Eigen::Vector2d &point) const {
        const float64_t dx =
            (point.x() > mid_x_ ? (point.x() - min_x_) : (point.x() - max_x_));
        const float64_t dy =
            (point.y() > mid_y_ ? (point.y() - min_y_) : (point.y() - max_y_));
        return std::hypot(dx, dy);
    }

    void GetAllObjects(std::vector<ObjectConstPtr> *result_objects) const {
        result_objects->insert(result_objects->end(),
                               objects_sorted_by_min_.begin(),
                               objects_sorted_by_min_.end());
        if (left_subnode_ != nullptr) {
            left_subnode_->GetAllObjects(result_objects);
        }
        if (right_subnode_ != nullptr) {
            right_subnode_->GetAllObjects(result_objects);
        }
    }
    void GetObjectsInternal(const Eigen::Vector2d &point,
                            float64_t distance,
                            std::vector<ObjectConstPtr> *result_objects) const {
        if (LowerDistanceToPoint(point) > distance) {
            return;
        }
        if (UpperDistanceToPoint(point) <= distance) {
            GetAllObjects(result_objects);
            return;
        }
        const float64_t pvalue =
            (partition_ == PARTITION_X ? point.x() : point.y());
        if (pvalue < partition_position_) {
            const float64_t limit = pvalue + distance;
            for (int i = 0; i < num_objects_; ++i) {
                if (objects_sorted_by_min_bound_[i] > limit) {
                    break;
                }
                ObjectConstPtr object = objects_sorted_by_min_[i];
                if (object->GetPoly2d().DistanceTo(point) <= distance) {
                    result_objects->push_back(object);
                }
            }
        } else {
            const float64_t limit = pvalue - distance;
            for (int i = 0; i < num_objects_; ++i) {
                if (objects_sorted_by_max_bound_[i] < limit) {
                    break;
                }
                ObjectConstPtr object = objects_sorted_by_max_[i];
                if (object->GetPoly2d().DistanceTo(point) <= distance) {
                    result_objects->push_back(object);
                }
            }
        }
        if (left_subnode_ != nullptr) {
            left_subnode_->GetObjectsInternal(point, distance, result_objects);
        }
        if (right_subnode_ != nullptr) {
            right_subnode_->GetObjectsInternal(point, distance, result_objects);
        }
    }

    void GetNearestObjectInternal(const Eigen::Vector2d &point,
                                  ObjectConstPtr *nearest_object,
                                  float64_t *min_distance) const {
        if (LowerDistanceToPoint(point) >= *min_distance - kMathEpsilon) {
            return;
        }
        const float64_t pvalue =
            (partition_ == PARTITION_X ? point.x() : point.y());
        const bool search_left_first = (pvalue < partition_position_);
        if (search_left_first) {
            if (left_subnode_ != nullptr) {
                left_subnode_->GetNearestObjectInternal(point, nearest_object,
                                                        min_distance);
            }
        } else {
            if (right_subnode_ != nullptr) {
                right_subnode_->GetNearestObjectInternal(point, nearest_object,
                                                         min_distance);
            }
        }
        if (*min_distance <= kMathEpsilon) {
            return;
        }

        if (search_left_first) {
            for (int i = 0; i < num_objects_; ++i) {
                const float64_t bound = objects_sorted_by_min_bound_[i];
                if (bound > pvalue &&
                    std::abs(bound - pvalue) > *min_distance) {
                    break;
                }
                ObjectConstPtr object = objects_sorted_by_min_[i];
                const float64_t distance =
                    object->GetPoly2d().DistanceTo(point);
                if (distance < *min_distance) {
                    *min_distance = distance;
                    *nearest_object = object;
                }
            }
        } else {
            for (int i = 0; i < num_objects_; ++i) {
                const float64_t bound = objects_sorted_by_max_bound_[i];
                if (bound < pvalue &&
                    std::abs(bound - pvalue) > *min_distance) {
                    break;
                }
                ObjectConstPtr object = objects_sorted_by_max_[i];
                const float64_t distance =
                    object->GetPoly2d().DistanceTo(point);
                if (distance < *min_distance) {
                    *min_distance = distance;
                    *nearest_object = object;
                }
            }
        }
        if (*min_distance <= kMathEpsilon) {
            return;
        }
        if (search_left_first) {
            if (right_subnode_ != nullptr) {
                right_subnode_->GetNearestObjectInternal(point, nearest_object,
                                                         min_distance);
            }
        } else {
            if (left_subnode_ != nullptr) {
                left_subnode_->GetNearestObjectInternal(point, nearest_object,
                                                        min_distance);
            }
        }
    }

    void ComputeBoundary(const std::vector<ObjectConstPtr> &objects) {
        min_x_ = std::numeric_limits<float64_t>::infinity();
        min_y_ = std::numeric_limits<float64_t>::infinity();
        max_x_ = -std::numeric_limits<float64_t>::infinity();
        max_y_ = -std::numeric_limits<float64_t>::infinity();
        for (ObjectConstPtr object : objects) {
            min_x_ = std::fmin(min_x_, object->GetPoly2d().MinX());
            max_x_ = std::fmax(max_x_, object->GetPoly2d().MaxX());
            min_y_ = std::fmin(min_y_, object->GetPoly2d().MinY());
            max_y_ = std::fmax(max_y_, object->GetPoly2d().MaxY());
        }
        mid_x_ = (min_x_ + max_x_) / 2.0;
        mid_y_ = (min_y_ + max_y_) / 2.0;
        if (!std::isinf(max_x_) && !std::isinf(max_y_) && !std::isinf(min_x_) &&
            !std::isinf(min_y_)) {
        } else {
            perror("the provided object box size is infinity\n");
        }
    }

    void ComputePartition() {
        if (max_x_ - min_x_ >= max_y_ - min_y_) {
            partition_ = PARTITION_X;
            partition_position_ = (min_x_ + max_x_) / 2.0;
        } else {
            partition_ = PARTITION_Y;
            partition_position_ = (min_y_ + max_y_) / 2.0;
        }
    }

    void PartitionObjects(const std::vector<ObjectConstPtr> &objects,
                          std::vector<ObjectConstPtr> *left_subnode_objects,
                          std::vector<ObjectConstPtr> *right_subnode_objects) {
        left_subnode_objects->clear();
        right_subnode_objects->clear();
        std::vector<ObjectConstPtr> other_objects;
        if (partition_ == PARTITION_X) {
            for (ObjectConstPtr object : objects) {
                if (object->GetPoly2d().MaxX() <= partition_position_) {
                    left_subnode_objects->push_back(object);
                } else if (object->GetPoly2d().MinX() >= partition_position_) {
                    right_subnode_objects->push_back(object);
                } else {
                    other_objects.push_back(object);
                }
            }
        } else {
            for (ObjectConstPtr object : objects) {
                if (object->GetPoly2d().MaxY() <= partition_position_) {
                    left_subnode_objects->push_back(object);
                } else if (object->GetPoly2d().MinY() >= partition_position_) {
                    right_subnode_objects->push_back(object);
                } else {
                    other_objects.push_back(object);
                }
            }
        }
        InitObjects(other_objects);
    }

 private:
    int num_objects_ = 0;
    std::vector<ObjectConstPtr> objects_sorted_by_min_;
    std::vector<ObjectConstPtr> objects_sorted_by_max_;
    std::vector<float64_t> objects_sorted_by_min_bound_;
    std::vector<float64_t> objects_sorted_by_max_bound_;
    int depth_ = 0;

    // Boundary
    float64_t min_x_ = 0.0;
    float64_t max_x_ = 0.0;
    float64_t min_y_ = 0.0;
    float64_t max_y_ = 0.0;
    float64_t mid_x_ = 0.0;
    float64_t mid_y_ = 0.0;

    enum Partition {
        PARTITION_X = 1,
        PARTITION_Y = 2,
    };
    Partition partition_ = PARTITION_X;
    float64_t partition_position_ = 0.0;

    std::unique_ptr<BoxKDTree2dNode<Object>> left_subnode_ = nullptr;
    std::unique_ptr<BoxKDTree2dNode<Object>> right_subnode_ = nullptr;
};

/**
 * @class BoxKDTree2d
 * @brief The class of KD-tree of Aligned Axis Bounding Box.
 */
template <class Object>
class BoxKDTree2d {
 public:
    using ObjectPtr = std::shared_ptr<Object>;
    using ObjectConstPtr = std::shared_ptr<const Object>;
    /**
     * @brief Contructor which takes a vector of objects and parameters.
     * @param params Parameters to build the KD-tree.
     */
    BoxKDTree2d(const std::unordered_map<std::string, ObjectPtr> &objects,
                const BoxKDTreeParams &params) {
        for (const auto &object : objects) {
            objects_.push_back(object.second);
        }
        if (!objects_.empty()) {
            root_.reset(new BoxKDTree2dNode<Object>(objects_, params, 0));
        }
    }

    /**
     * @brief Get the nearest object to a target point.
     * @param point The target point. Search it's nearest object.
     * @return The nearest object to the target point.
     */
    void GetNearestObject(const Eigen::Vector2d &point,
                          ObjectConstPtr *object,
                          float64_t *dis) const {
        if (object == nullptr || dis == nullptr) {
            AD_LERROR(HDMAP) << "dis or object is nullptr";
            return;
        }
        if (root_ == nullptr) {
            AD_FDEBUG(HDMAP) << "kdtree root is nullptr";
            return;
        }
        *dis = std::numeric_limits<float64_t>::max();
        root_->GetNearestObject(point, object, dis);
    }
    /**
     * @brief Get objects within a distance to a point.
     * @param point The center point of the range to search objects.
     * @param distance The radius of the range to search objects.
     * @return All objects within the specified distance to the specified
     point.
     */
    void GetObjects(const Eigen::Vector2d &point,
                    float64_t distance,
                    std::vector<ObjectConstPtr> *objects) const {
        if (root_ != nullptr) {
            root_->GetObjects(point, distance, objects);
        } else {
            AD_FDEBUG(HDMAP) << "root is nullptr";
        }
    }

 private:
    std::unique_ptr<BoxKDTree2dNode<Object>> root_ = nullptr;
    std::vector<ObjectConstPtr> objects_;
};

}  // namespace hdm
}  // namespace saffron
