#pragma once
#include <string>
#include <vector>
#include <memory>
#include <unordered_map>
#include <limits>
#include <utility>

#include "map/data_type/hdmap_data.hpp"
#include "map/common/status.hpp"
#include "map/common/error.hpp"
#include "map/semantic_map/line.hpp"
#include "map/math_tools/polygon2d.hpp"

namespace saffron {
namespace hdm {

/*************************** Lane ********************************/
class Lane {
 public:
    using ConstPtr = std::shared_ptr<const Lane>;
    using Ptr = std::shared_ptr<Lane>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    Lane() = default;
    explicit Lane(LaneData &&lane_data);
    explicit Lane(const LaneData &lane_data);
    const LaneData &GetData() const;
    const std::string &GetId() const;
    const std::string &GetRoadId() const;
    const std::string &GetSectionId() const;
    float64_t GetLength() const;
    const std::vector<std::string> &GetOverlapIds() const;
    const std::string &GetLeftLineId() const;
    const std::string &GetRightLineId() const;
    const std::string &GetCenterLineId() const;
    float64_t GetMaxSpeed() const;
    float64_t GetMinSpeed() const;
    const std::vector<float64_t> &GetWidths() const;
    float64_t GetWidth(const PointENU_t &position,
                       const Line::ConstPtr &center_line) const;
    const std::vector<std::string> &GetSuccessorLinkIds() const;
    const std::vector<std::string> &GetPredecessorLinkIds() const;
    const NeighborLane &GetLeftNeighbor() const;
    const NeighborLane &GetRightNeighbor() const;
    const std::string &GetLeftNeighborId() const;
    const std::string &GetRightNeighborId() const;
    bool IsLeftNeighborExist() const;
    bool IsRightNeighborExist() const;
    bool IsLeftNeighborForward() const;
    bool LeftNeighborCanChange() const;
    bool IsRightNeighborForward() const;
    bool RightNeighborCanChange() const;
    uint16_t GetTurnType() const;
    bool GoStraight() const;
    bool TurnLeft() const;
    bool TurnRight() const;
    bool LeftAround() const;
    bool RightAround() const;
    bool TurnLeftRight() const;
    bool TurnLeftWithAround() const;
    bool TurnRightWithAround() const;
    bool StraightLeft() const;
    bool StraightRight() const;
    bool StraightLeftAround() const;
    bool StraightRightAround() const;
    bool StraightLeftWithAround() const;
    bool StraightRightWithAround() const;
    bool CanTurnLeft() const;
    bool CanTurnRight() const;
    bool CanLeftAround() const;
    bool CanRightAround() const;
    bool CanGoStraight() const;
    bool Free2Go() const;
    LaneType GetLaneType() const;
    bool IsWaitingZone() const;
    bool IsDrivable() const;

    const Polygon2d &GetPoly2d() const;
    const Point3D_t &GetEndPoint() const;
    const Point3D_t &GetBeginPoint() const;
    mapStatus_t SetPolygon(const Line::ConstPtr &left_line,
                           const Line::ConstPtr &right_line);
    bool IsPointIn(const Point3D_t &position) const;
    float64_t Distance2Point(const Point3D_t &position) const;
    std::vector<std::string> GetJunctionIds() const;
    std::vector<std::string> GetStopLineIds() const;
    std::vector<std::string> GetGroupTLIds() const;

 public:
    /// TODO(chenlongquan): hardcode for dpc, use better way in the future
    mapStatus_t SetDangerPoints(const Point3D_t &start_danger_pt,
                                const Point3D_t &end_danger_pt);
    bool HasDangerArea() const;
    std::pair<Point3D_t, Point3D_t> GetDangerPoints() const;

 private:
    LaneData lane_data_;
    Polygon2d polygon_2d_;
    Point3D_t end_point_{};
    Point3D_t begin_point_{};

    /// TODO(chenlongquan): hardcode for dpc, use better way in the future
    Point3D_t start_danger_pt_{};
    Point3D_t end_danger_pt_{};
};

/*************************** LaneLink ********************************/
class LaneLink {
 public:
    using ConstPtr = std::shared_ptr<const LaneLink>;
    using Ptr = std::shared_ptr<LaneLink>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    LaneLink() = default;
    explicit LaneLink(LaneLinkData &&lane_link_data);
    explicit LaneLink(const LaneLinkData &lane_link_data);
    const LaneLinkData &GetData() const;
    const std::string &GetId() const;
    const std::string &GetFromLaneId() const;
    const std::string &GetToLaneId() const;
    const std::string &GetRefLineId() const;
    bool IsContinuity() const;
    TurnLink GetTurnLink() const;

 private:
    LaneLinkData lane_link_data_;
};
}  // namespace hdm

}  // namespace saffron
