#pragma once
#include <string>
#include <vector>
#include <memory>
#include <unordered_map>
#include <utility>

#include "map/data_type/hdmap_data.hpp"
#include "map/math_tools/polyline2d.hpp"
#include "map/math_tools/polygon2d.hpp"

namespace saffron {
namespace hdm {
template <typename DataType, typename PolyType>
class Object {
 public:
    Object() = default;
    explicit Object(DataType &&data) : data_(std::move(data)) { SetPoly(); }
    explicit Object(const DataType &data) : data_(data) { SetPoly(); }
    const DataType &GetData() const { return data_; }
    const std::string &GetId() const { return data_.id; }
    const std::vector<Point3D_t> &GetGeometry() const { return data_.geometry; }
    const PolyType &GetPoly2d() const { return poly_2d; }
    bool IsPointIn(const Point3D_t &position) const {
        return poly_2d.IsPointIn(Eigen::Vector2d(position.x, position.y));
    }
    float64_t Distance2Point(const Point3D_t &position) const {
        return poly_2d.DistanceTo(Eigen::Vector2d(position.x, position.y));
    }
    float64_t HeadingFromPoint2Poly(const Point3D_t &position) const {
        return poly_2d.HeadingFrom(Eigen::Vector2d(position.x, position.y));
    }

    float64_t GetCenterX() { return poly_2d.CenterX(); }
    float64_t GetCenterY() { return poly_2d.CenterY(); }

 protected:
    void SetPoly() {
        std::vector<Eigen::Vector2d> poly_data;
        for (const auto &geo : data_.geometry) {
            poly_data.emplace_back(geo.x, geo.y);
        }
        poly_2d = PolyType(poly_data);
    }
    DataType data_;
    PolyType poly_2d;
};
/*************************** guide sign ********************************/
class GuideSign : public Object<GuideSignData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const GuideSign>;
    using Ptr = std::shared_ptr<GuideSign>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    GuideSign() = default;
    explicit GuideSign(GuideSignData &&data);
    explicit GuideSign(const GuideSignData &data);
    const std::string &GetType() const;
};

/*************************** traffic light ********************************/
class TrafficLight : public Object<TrafficLightData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const TrafficLight>;
    using Ptr = std::shared_ptr<TrafficLight>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    TrafficLight() = default;
    explicit TrafficLight(TrafficLightData &&data);
    explicit TrafficLight(const TrafficLightData &data);
    const std::string &GetType() const;
};

/*************************** obstacle ********************************/
class Obstacle : public Object<ObjectData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const Obstacle>;
    using Ptr = std::shared_ptr<Obstacle>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    Obstacle() = default;
    explicit Obstacle(ObjectData &&data);
    explicit Obstacle(const ObjectData &data);
};

/*************************** side walk ********************************/
class Sidewalk : public Object<ObjectData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const Sidewalk>;
    using Ptr = std::shared_ptr<Sidewalk>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    Sidewalk() = default;
    explicit Sidewalk(ObjectData &&data);
    explicit Sidewalk(const ObjectData &data);
};

/*************************** stop line ********************************/
class StopLine : public Object<ObjectData, Polyline2d> {
 public:
    using ConstPtr = std::shared_ptr<const StopLine>;
    using Ptr = std::shared_ptr<StopLine>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    StopLine() = default;
    explicit StopLine(ObjectData &&data);
    explicit StopLine(const ObjectData &data);
};

/*************************** pole ********************************/
class Pole : public Object<ObjectData, Polyline2d> {
 public:
    using ConstPtr = std::shared_ptr<const Pole>;
    using Ptr = std::shared_ptr<Pole>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    Pole() = default;
    explicit Pole(ObjectData &&data);
    explicit Pole(const ObjectData &data);
};

/*************************** road marker ********************************/
class RoadMarker : public Object<ObjectData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const RoadMarker>;
    using Ptr = std::shared_ptr<RoadMarker>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    RoadMarker() = default;
    explicit RoadMarker(ObjectData &&data);
    explicit RoadMarker(const ObjectData &data);
};

/*************************** cross walk ********************************/
class Crosswalk : public Object<CrosswalkData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const Crosswalk>;
    using Ptr = std::shared_ptr<Crosswalk>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    Crosswalk() = default;
    explicit Crosswalk(CrosswalkData &&data);
    explicit Crosswalk(const CrosswalkData &data);
    const std::vector<std::string> &GetContainIds() const;
};

/*************************** park area ********************************/
class ParkArea : public Object<ObjectData, Polygon2d> {
 public:
    using ConstPtr = std::shared_ptr<const ParkArea>;
    using Ptr = std::shared_ptr<ParkArea>;
    using Table = std::unordered_map<std::string, Ptr>;
    using Vector = std::vector<ConstPtr>;

 public:
    ParkArea() = default;
    explicit ParkArea(ObjectData &&data);
    explicit ParkArea(const ObjectData &data);
};

}  // namespace hdm
}  // namespace saffron
